﻿using XF.Domain.Dependency;
using XF.Domain.Exceptions;

namespace Hotline.Settings;

public class SystemDomainService : ISystemDomainService, IScopeDependency
{
    private readonly ISystemOrganizeRepository _organizeRepository;
    private readonly ISysDicDataRepository _sysDicDataRepository;

    public SystemDomainService(
        ISystemOrganizeRepository organizeRepository,
        ISysDicDataRepository sysDicDataRepository)
    {
        _organizeRepository = organizeRepository;
        _sysDicDataRepository = sysDicDataRepository;
    }

    public async Task<IEnumerable<KeyValuePair<int, string>>> QueryOrgLevelOptionsAsync(CancellationToken cancellationToken)
    {
        var max = await _organizeRepository.Queryable()
            .MaxAsync(d => d.OrgLevel);

        var rsp = new List<KeyValuePair<int, string>>();
        for (int i = 1; i <= max; i++)
        {
            rsp.Add(new KeyValuePair<int, string>(i, $"{i}级部门"));
        }

        return rsp;
    }

    public async Task<IEnumerable<KeyValuePair<string, string>>> QueryOrgLevelStringOptionsAsync(CancellationToken cancellationToken)
    {
        var rsp = await QueryOrgLevelOptionsAsync(cancellationToken);
        return rsp.Select(d => new KeyValuePair<string, string>(d.Key.ToString(), d.Value));
    }

    public async Task<IReadOnlyList<SysDicData>> GetSysDicDataByCodeAsync(string code, CancellationToken cancellationToken = default)
    {
        return await _sysDicDataRepository.Queryable().Where(x => x.DicTypeCode == code).ToTreeAsync(x => x.Children, x => x.ParentId, "");
    }

    /// <summary>
    /// 生成新的部门编码
    /// </summary>
    /// <param name="parentId"></param>
    /// <param name="cancellationToken"></param>
    /// <returns></returns>
    public async Task<string> GenerateNewOrgCodeAsync(string parentId, CancellationToken cancellationToken)
    {
        var org = await _organizeRepository.GetAsync(parentId, cancellationToken);
        if (org == null)
            throw new UserFriendlyException("无效上级部门编码");
        var count = await _organizeRepository.CountAsync(d => d.ParentId == parentId, cancellationToken);
        var current = (count + 1).ToString("000");
        var code = $"{org.OrgCode}{current}";
        var exists = await _organizeRepository.AnyAsync(d => d.OrgCode == code, cancellationToken);
        while (exists)
        {
            return await GenerateNewOrgCodeAsync(parentId, cancellationToken);
        }
        return code;
    }
}